;+ 
; Aim:   solve equation  dy/dt = div( D(S(grad(us))).grad(y) ),  y(0) = u,  y(t+T) = y(t) + T*dy/dt
; Use:   u=nonlineardiff(2Dimage,signma,rho,m=m,dt=dt, ndt=ndt,mode=mode,scheme=scheme)
; Algorithm: depends on scheme and mode
; input:
; u: 2D image
; sigma: 0.5 (default) the sigma of gaussian function for regulation (smoothing) before calculating the gradient
; rho: 1.0 (default for coherence enhancing) or 0.0 (default for edge enhancing), smoothing for structure tensor 
; m: 4 (default) parameter differencing vertical and parallel to gradient  diffusion
; dt: 0.01(default) time steps for diffusion
; ndt: 10 (default) number of steps of diffusion
; mode: 'eed' for edge enhanced diffusion filtering 
; 	    'ced' for coherence enhanced diffusion filtering
;            parameters for the diffusiontensor has to be modified by hand
; scheme: discretization scheme for solved the equation
; 	'standard': Euler explicit method
; 	'non_negative': non_negative method
; 	'novel': novel method
; 	'implicit': implict method
;         'rotation_invariant': rotation_invariant method
; output: updated image after non-linear diffusion filtering
; Dependence:
; Contact:   bugs and comments to:  
; 		Ding Yuan (Ding.Yuan@warwick.ac.uk)
; 		CFSA, Physics, Warwick University 
; Acknowledgement: 
; 	The codes are mainly based on the nonlinear diffusion package
; 	  uploaded to the Matlab file exchange system
; 	  1. Image Edgee Enhancing Coherence Filter Toolbox
; 	  by Dirk-Jan Kroon (September 2009)
; 	  http://www.mathworks.com/matlabcentral/fileexchange/25449-image-edge-enhancing-coherence-filter-toolbox
; 	  2.  Nonlinear Diffusion Toolbox by Frederico D'Almeida (March 2004)
; 	  http://www.mathworks.com/matlabcentral/fileexchange/3710-nonlinear-diffusion-toolbox
; History: 
; 	    10 Feb 2011 - First version
;            
;-
function gsderiv, u, sigma,order=order
; + 
; Aim: Calculates the derivate of the gaussian smoothed image.
; Use:   us= GSDERIV( u, sigma, order=order)
;    calculates the "order"-th derivates of image u
;    Prior to derivating the image "u" is smoothed by a ''sigma'' 
;    standard deviation 2D gaussian.
;    order=0, return gaussian smoothed images 
;    order=1, return us[nx,ny,2]=[[[ux[nx,ny]],[[uy[nx,ny]]]
;    u must have two dimensions.
; Algorithm:   
; Dependence: 
; exist.pro availabe in solar soft
; gaussian_kernel.pro
; gradient.pro
; Contact: Ding Yuan (Ding.Yuan@warwick.ac.uk)
; 		CFSA, Physics, Warwick University 
; History: 09-Feb-2011 first version   
; -

; Verifying inputs
if n_params() lt 2 then message,'Please input at least two parameters'
if ~exist(u) then on_error,2
if ~exist(sigma) then sigma=0
if size(u,/n_dimensions) ne 2 then message, 'Use 2D images.'
u=double(u) 
if ~exist(order) then order=1 else  order=fix(order) ; default order 1
if ~in_range(order,[0,1]) then message, 'Please input order as option of 0,1, the default is 1 '
kernel=gaussian_kernel(sigma) ; Calculate smoothing kernel
; Convolution image u kernel (smoothing)
if n_elements(kernel) gt 1 then us = convol(u,kernel,/edge_truncate,/center) else us=u
; Calculate image derivates
case  order of 
      0: return, us
      1: begin
	    y= gradient(us,/vector) 
	    return,y
	 end
      else: on_error,2
endcase
end
;--------------End ---------------- Derivatives of Gaussian Smoothed Image------------
;****** Structure Tensor ****************************************
function structuretensor, ux,uy,rho
; + 
; Aim:  Calculate 2D structure tensor for non-linear diffusion
;         Jp ( grad(u) ) = conv ( Kp  , grad(u) * grad(u)^T ) 
;         The 2x2 JP matrix is called structure tensor (second-moment matrix,
;          scatter matrix, Forstner interest operator)
;           J. Weickert "Coherence-Enhancing Shock filters"
; Use:   called in nonlineardiff.pro
; Algorithm:
; Dependence: exist.pro (solarsoft)
; Contact: Ding Yuan (Ding.Yuan@warwick.ac.uk)
; 		CFSA, Physics, Warwick University 
; History: 09-Feb-2011
; -
if n_params() lt 2 then on_error,2
if ~exist(ux) then on_error,2
if ~exist(uy) then on_error,2
if ~exist(rho) then begin
rho=0
message,'On input of rho is given, 0 is assigned as default',/info
endif
; J(grad u_sigma)
Jxx = ux^2
Jxy = ux*uy
Jyy = uy^2

kernel=gaussian_kernel(rho)
if n_elements(kernel) gt 1 then begin
; Do the gaussian smoothing of the structure tensor 
; Note derivation and convolation are commutable
Jxx = convol(Jxx,kernel,/edge_truncate,/center)
Jxy = convol(Jxy,kernel,/edge_truncate,/center)
Jyy = convol(Jyy,kernel,/edge_truncate,/center)
endif
return,[[[Jxx]],[[Jxy]],[[Jyy]]]
end
;-------End ---------------- Structure Tensor------------
;********* eivenvector ******************************
function eigenvector, Jxx,Jxy,Jyy
; This function computes the eigenvectors and eigen values of the 2D image
; Hessian
;
; eigen=eigenvectors(Jxx,Jxy,Jyy)
;
; Contact: Ding Yuan (Ding.Yuan@warwick.ac.uk)
;		CFSA, Physics, Warwick University 
;History:  09-Feb-2011

; Compute the eigenvectors of JJ
tmp = sqrt((Jxx - Jyy)^2 + 4*Jxy^2)
v1x = 2*Jxy & v1y = Jyy - Jxx + tmp
; help,v1x,v1y
; Normalize
mag = sqrt(v1x^2 + v1y^2) 
valid=where(mag gt 1e-7) 
v1x(valid) = v1x(valid)/mag(valid)
v1y(valid) = v1y(valid)/mag(valid)

; The eigenvectors are orthogonal
v2x = -v1y; 
v2y = v1x;
; help,v2x,v2y
; Compute the eigenvalues
mu1 = 0.5*(Jxx + Jyy + tmp);
mu2 = 0.5*(Jxx + Jyy - tmp);
A={eigen,values: [[[mu1]],[[mu2]]],vectors:[[[v1x]],[[v1y]],[[v2x]],[[v2y]]]}

return,A
end

;-------End----- Eigenvector---------------------------------------
;************ diffusiontensor****************************
function diffusiontensor, eigen,gradA=gradA,m=m,lambda_e=lambda_e,mode=mode
; Construct the diffusion tensor D = [[ a, b],$
; 							  [ b, c]] 
;
;
; 
; Contact: Ding Yuan (Ding.Yuan@warwick.ac.uk)
; 		CFSA, Physics, Warwick University 
; History:
;-

mu1=reform(eigen.values[*,*,0]) & mu2= reform(eigen.values[*,*,1])
v1x=reform(eigen.vectors[*,*,0]) & v1y=reform(eigen.vectors[*,*,1])
v2x=reform(eigen.vectors[*,*,2]) & v2y=reform(eigen.vectors[*,*,3])

if ~exist(mode) then mode='eed'
case mode of 
	'eed': begin
		    if ~exist(m) then m=4
		    if ~exist(lambda_e) then lambda_e=2.0		    
		    ; 1-exp(-x)-m*x*exp(-x)=0, solved by matlab
		    index_m=findgen(38)*0.5+1.5 
                    value=[ 0.7627,    1.2564,    1.6188,   1.9038,    2.1381,    2.3367,    2.5087,    2.6604,    2.7959,    2.9183,    3.0298,    3.1323,    3.2269,$
                             3.3149,    3.3970,    3.4740,    3.5465,    3.6150,    3.6798,    3.7413,    3.8000,    3.8559,    3.9094,    3.9606,    4.0097,    4.0569,$
	                     4.1024,    4.1462,    4.1885,    4.2293,    4.2688,    4.3070,    4.3441,    4.3800,    4.4149,    4.4488,    4.4818,   4.5139]
                    cm0=interpol(value,index_m,m,/spline) ; find Cm by spline interpolation	     
;		    cm=reform(cm0)  ; 1*1 array to scalar
		    cm=cm0[0] ; reform converts 1*1 into 1*1
;		    help,cm0,cm
		    lambda2 = 1 - exp(-cm/(gradA/lambda_e^2)^m) 
;		    help,lambda2
		    lower=where(gradA lt 1e-7)
		    if product(lower) ge 0 then lambda2[lower]=1 
		    lambda1 = 1.
		end
	'ced': begin
		     if ~exist(m) then m=1
		     if ~exist(alpha) then  alpha=0.001
		     if ~exist(C) then C=1.    
		    di=(mu1-mu2) 
		    valid=where((di lt 1e-7) and (di gt -1e-17))
		    if product(valid) ge 0 then di(where((di lt 1e-7) and (di gt -1e-17)))=1e-7
		    lambda2 = alpha + (1 - alpha)*exp(-C/di^(2*m)) 
		    lambda1 = alpha
		end
	else: message,'Unknown mode inserted, please input mode = ''eed'' or ''ced''.'
endcase

Dxx = lambda1*v1x^2   + lambda2*v2x^2
Dxy = lambda1*v1x*v1y + lambda2*v2x*v2y
Dyy = lambda1*v1y^2   + lambda2*v2y^2


return,[[[Dxx]],[[Dxy]],[[Dyy]]]
end
;-------End--------------- diffusion tensor------------------------------------

;************************ Standard Scheme********************************************
function standard, u, Dxx,Dxy,Dyy,dt
if ~exist(u) then on_error,2
sz=size(u)
if sz[0] ne 2 then on_error,2
N1=sz[1] & N2=sz[2]
if ~exist(Dxx) then on_error,2
sz=size(Dxx)
if sz[1] ne N1 or sz[2] ne N2 then on_error,2
if ~exist(Dxy) then on_error,2
sz=size(Dxy)
if sz[1] ne N1 or sz[2] ne N2 then on_error,2
if ~exist(Dyy) then on_error,2
sz=size(Dyy)
if sz[1] ne N1 or sz[2] ne N2 then on_error,2
a=Dxx & b=Dxy & c=Dyy 
A1=(1./4.)*(roll(b,[1,0])-roll(b,[0,-1])) 
A2=(1./2.)*(roll(c,[0,-1])+c)
A3=(1./4.)*(roll(b,[-1,0])+roll(b,[0,-1])) 
A4=(1./2.)*(roll(a,[1,0])+a)
;A5 = -(1/2)*(a[nx,*]+2*a+a[px,*]) -(1/2)*(c[*,ny]+2*c+c([*,py]);
A6=(1./2.)*(roll(a,[-1,0])+a)
A7=(1./4.)*(roll(b,[1,0])+roll(b,[0,1]))
A8=(1./2.)*(roll(c,[0,1])+c)
A9=(1./4.)*(roll(b,[-1,0]) -roll(b,[0,1])) 
du=A1*(roll(u,[1,-1])-u)$
    +A2*(roll(u,[0,-1])-u)$
    +A3*(roll(u,[-1,-1])-u)$
    +A4*(roll(u,[1,0])-u)$
    +A6*(roll(u,[-1,0])-u)$
    +A7*(roll(u,[1,1])-u)$
    + A8*(roll(u,[0,1])-u)$
    +A9*(roll(u,[-1,1])-u)  
return, u+du*dt
end 
;---------- End ----------- Standard scheme-----------------------
;************ Non_negative*************************
function non_negative,u,Dxx,Dxy,Dyy,dt
if ~exist(u) then on_error,2
sz=size(u)
if sz[0] ne 2 then on_error,2
N1=sz[1] & N2=sz[2]
if ~exist(Dxx) then on_error,2
sz=size(Dxx)
if sz[1] ne N1 or sz[2] ne N2 then on_error,2
if ~exist(Dxy) then on_error,2
sz=size(Dxy)
if sz[1] ne N1 or sz[2] ne N2 then on_error,2
if ~exist(Dyy) then on_error,2
sz=size(Dyy)
if sz[1] ne N1 or sz[2] ne N2 then on_error,2

a=Dxx & b=Dxy& c=Dyy
A1 = (1./4.)*((abs(roll(b,[1, -1]))-roll(b,[1,-1])) + (abs(b)-b))
A2 = (1./2.)*((roll(c,[0, -1])+c) -(abs(roll(b,[0,-1]))+abs(b)))
A3 = (1./4.)*((abs(roll(b,[-1, -1]))+roll(b,[-1,-1])) + (abs(b)+b))
A4 = (1./2.)*( (roll(a,[1,0])+a) -(abs(roll(b,[1,0]))+abs(b)))
;A5 =  - (a(nx,:) + 2*a+a(px,:)) ...
;	  - (abs(b(nx,py))-b(nx,py)+abs(b(px,py))-b(px,py)) ...
;	  - (abs(b(nx,ny))-b(nx,ny)+abs(b(px,ny))-b(px,ny)) ...
;	  + (abs(b(nx,:))+abs(b(px,:))+abs(b(:,ny))+abs(b(:,py))+2*abs(b)) ...
;	  - (c(:,ny) + 2*c+c(:,py));
A6 = (1./2.)*( (roll(a,[-1,0])+a) -(abs(roll(b,[-1,0]))+abs(b)));
A7 = (1./4.)*((abs(roll(b,[1, 1]))+roll(b,[1,1])) + (abs(b)+b));
A8 = (1./2.)*( (roll(c,[0, 1])+c) -(abs(roll(b,[0,1]))+abs(b)));
A9 = (1./4.)*((abs(roll(b,[-1, 1]))-roll(b,[-1,1])) + (abs(b)-b));
du=A1*(roll(u,[1,-1]) -u)$
    +A2*(roll(u,[0, -1]) -u)$
    +A3*(roll(u,[-1,-1]) -u)$
    +A4*(roll(u,[1,0] ) -u)$    
    +A6*(roll(u,[-1,0]) -u)$
    +A7*(roll(u,[1,1]) -u)$
    +A8*(roll(u,[0, 1]) -u)$
    +A9*(roll(u,[-1,1]) -u)
return,du*dt+u
end
;------------ end non_negative --------------------------------------
;************* implicit ******************************
function implicit,u,Dxx,Dxy,Dyy,dt
if ~exist(u) then on_error,2
sz=size(u)
if sz[0] ne 2 then on_error,2
N1=sz[1] & N2=sz[2]
if ~exist(Dxx) then on_error,2
sz=size(Dxx)
if sz[1] ne N1 or sz[2] ne N2 then on_error,2
if ~exist(Dxy) then on_error,2
sz=size(Dxy)
if sz[1] ne N1 or sz[2] ne N2 then on_error,2
if ~exist(Dyy) then on_error,2
sz=size(Dyy)
if sz[1] ne N1 or sz[2] ne N2 then on_error,2

Dul = (roll(Dxy, [1,1])>0) + (Dxy>0)
Ddr = (roll(Dxy, [-1,-1])>0)+ (Dxy>0)
Ddl = (-roll(Dxy,[-1,1])>0) + (-Dxy>0)
Dur = (-roll(Dxy,[1,-1])>0) +(-Dxy>0)

; Normal neighbor intensity updates 
Dml = (roll(Dyy,[0,1]) + Dyy) - (abs(roll(Dxy,[0,1])) + abs(Dxy))
Dmr = (roll(Dyy,[0,-1]) + Dyy) - (abs(roll(Dxy,[0,-1])) + abs(Dxy))
Duc = (roll(Dxx,[1,0]) + Dxx) - (abs(roll(Dxy,[1,0])) + abs(Dxy))
Ddc = (roll(Dxx,[-1,0]) + Dxx) - (abs(roll(Dxy,[-1,0])) + abs(Dxy))
; Normalization term to preserve region average grey value
Dmc = -(roll(Dyy,[0,1]) + 2*Dyy + roll(Dyy,[0,-1]))$ 
    -(roll(Dxx, [1,0]) + 2*Dxx + roll(Dxx,[-1,0]))$  
    -((roll(Dxy,[1,1])>0) + (-roll(Dxy,[1,-1])>0))$
    -((-roll(Dxy,[-1,1])>0) + (roll(Dxy,[-1,-1])>0))$
    +(abs(roll(Dxy,[0,1])) + abs(roll(Dxy,[0,-1])) + abs(roll(Dxy,[-1,0])) + abs(roll(Dxy,[1,0])) + 2*abs(Dxy));
; Calculate Diffusion update 
; du=dblarr(N1,N2);
du=0.5*dt*(Dul*roll(u,[1,1])$
		  +Dml*roll(u,[0,1])$
		  +Ddl*roll(u,[-1,1])$
		  +Duc*roll(u,[1,0])$           
		  +Ddc*roll(u,[-1,0])$
		  +Dur*roll(u, [1,-1])$
		  +Dmr*roll(u, [0,-1])$
		  +Ddr*roll(u,[-1,-1]))
return, (u + du) / (1.-dt*0.5*Dmc)
end
;------------------- end of implicit --------------------------------
;*********** Novel scheme *****************************
function novel,u,Dxx,Dxy,Dyy,dt
if ~exist(u) then on_error,2
sz=size(u)
if sz[0] ne 2 then on_error,2
N1=sz[1] & N2=sz[2]
if ~exist(Dxx) then on_error,2
sz=size(Dxx)
if sz[1] ne N1 or sz[2] ne N2 then on_error,2
if ~exist(Dxy) then on_error,2
sz=size(Dxy)
if sz[1] ne N1 or sz[2] ne N2 then on_error,2
if ~exist(Dyy) then on_error,2
sz=size(Dyy)
if sz[1] ne N1 or sz[2] ne N2 then on_error,2

par=[0.00549691757211334,4.75686155670860e-10,$
	3.15405721902292e-11,0.00731109320628158,$
	1.40937549145842e-10,0.0876322157772825,$
	0.0256808495553998,5.87110587298283e-11,$
	0.171008417902939,3.80805359553021e-12,$
	9.86953381462523e-12,0.0231020787600445,$
	0.00638922328831119,0.0350184289706385]
Mxx =[[par[0],  par[1],  par[2],  par[1],  par[0]],$
	    [par[3],  par[4],  par[5],  par[4],  par[3]],$
	    [-par[6], -par[7], -par[8], -par[7], -par[6]],$
	    [par[3],  par[4],  par[5],  par[4],  par[3]],$
	    [par[0],  par[1],  par[2],  par[1],  par[0]]]
Myy=transpose(Mxx)
Mxy=[[  par[9],   par[10],    0,    -par[10], -par[9]],$
	  [par[10],   par[11],    0,    -par[11], -par[10]],$
	  [         0,            0,    0,              0,          0],$
	  [-par[10], -par[11],    0,    par[11],   par[10]],$
	  [  -par[9], -par[10],    0,    par[10],   par[9]]]

Mx= [[par[12],  par[13],  par[12]],$ 
	  [	0, 	    0, 	      0],$ 
	  [-par[12], -par[13], -par[12]]]
My= transpose(Mx)
ux=convol(u,Mx,/edge_truncate,/center)
uy=convol(u,My,/edge_truncate,/center)
div1=convol(Dxx,Mx,/edge_truncate,/center)+convol(Dxy,My,/edge_truncate,/center)
div2=convol(Dxy,Mx,/edge_truncate,/center)+convol(Dyy,My,/edge_truncate,/center)
du1= div1*ux+div2*uy 
uxx=convol(u,Mxx,/edge_truncate,/center)
uxy=convol(u,Mxy,/edge_truncate,/center)
uyy=convol(u,Myy,/edge_truncate,/center)
du2=uxx*Dxx+2*uxy*Dxy + uyy*Dyy
du=du1+du2 
u=u+du*dt 
infinite=where(~finite(u))
if product(infinite) ge 0 then u[infinite]=0
return,u
end
;------ end novel -----------------------------------------
;******** Rotation_invariant ****************************
; ;********************* Rotation Invariant Stencil Difference **************************
; function scharr, u, x=x,y=y
; u=double(u)
; sz=size(u)
; if sz[0] ne 2 then message, 'Only 2D works'
; 
; if keyword_set(x) then  $
; 	Stencil= [[3, 10, 3],$
; 		       [0,   0,  0],$
; 		       [-3, -10, -3]]/32.
; 
; if keyword_set(y) then $
;  Stencil= [[3,  0, -3],$
; 	        [10, 0, -10],$
; 	        [3,  0, -3]]/32.   
; 
; du=convol(u,stencil,/edge_truncate,/center)
; return, du
; end

function rotation_invariant,u,Dxx,Dxy,Dyy,dt
if ~exist(u) then on_error,2
sz=size(u)
if sz[0] ne 2 then on_error,2
N1=sz[1] & N2=sz[2]
if ~exist(Dxx) then on_error,2
sz=size(Dxx)
if sz[1] ne N1 or sz[2] ne N2 then on_error,2
if ~exist(Dxy) then on_error,2
sz=size(Dxy)
if sz[1] ne N1 or sz[2] ne N2 then on_error,2
if ~exist(Dyy) then on_error,2
sz=size(Dyy)
if sz[1] ne N1 or sz[2] ne N2 then on_error,2
xstencil= [[3, 10, 3],$
		[0,   0,  0],$
		[-3, -10, -3]]/32.
ystencil= [[3,  0, -3],$
	        [10, 0, -10],$
	        [3,  0, -3]]/32.  
ux=convol(u,xstencil,/edge_truncate,/center)
uy=convol(u,ystencil,/edge_truncate,/center)
; ux=scharr(u,/x)
; uy=scharr(u,/y)
; 3 : Calculate the flux components
j1 = Dxx * ux + Dxy *uy;
j2 = Dxy * ux + Dyy *uy
; 3.5 Flux on boundaries to zero
j1[0,*]=0 & j1[N1-1,*]=0 & j1[*,0]=0 & j1[*,N2-1]=0;
j2[0,*]=0 & j2[N1-1,*]=0 & j2[*,0]=0 & j2[*,N2-1]=0;
; 4 : Calculate ... by means of the optimized derivative filters
j1x=convol(j1,xstencil,/edge_truncate,/center)
j2y=convol(j2,ystencil,/edge_truncate,/center)
; du = scharr(j1,/x)+scharr(j2,/y);
du = j1x+j2y
; 5 : Update in an explicit way.
u=u+du*dt
return,u
end

;---- end rotation_invariant -----------------------------
;************************************************************************
;**************** Start of Main function *******************************
;************************************************************************
function nonlineardif, u, sigma, rho, m=m,dt=dt, ndt=ndt,mode=mode,scheme=scheme    
;+ 
; Aim:   solve equation  dy/dt = div( D(S(grad(us))).grad(y) ),  y(0) = u,  y(t+T) = y(t) + T*dy/dt
; Use:   u=nonlineardiff(2Dimage,signma,rho,m=m,dt=dt, ndt=ndt,mode=mode,scheme=scheme)
; Algorithm: depends on scheme and mode
; input:
; u: 2D image
; sigma: 0.5 (default) the sigma of gaussian function for regulation (smoothing) before calculating the gradient
; rho: 1.0 (default for coherence enhancing) or 0.0 (default for edge enhancing), smoothing for structure tensor 
; m: 4 (default) parameter differencing vertical and parallel to gradient  diffusion
; dt: 0.01(default) time steps for diffusion
; ndt: 10 (default) number of steps of diffusion
; mode: 'eed' for edge enhanced diffusion filtering
; 	    'ced' for coherence enhanced diffusion filtering
; scheme: discretization scheme for solved the equation
; 	'standard': Euler explicit method
; 	'non_negative': non_negative method
; 	'novel': novel method
; 	'implicit': implict method
;         'rotation_invariant': rotation_invariant method
; output: updated image after non-linear diffusion filtering
; Dependence:
; Contact:   bugs and comments to:  
; 		Ding Yuan (Ding.Yuan@warwick.ac.uk)
; 		CFSA, Physics, Warwick University 
; Acknowledgement: 
; 	The codes are mainly based on the nonlinear diffusion package
; 	  uploaded to the Matlab file exchange system
; 	  1. Image Edgee Enhancing Coherence Filter Toolbox
; 	  by Dirk-Jan Kroon (September 2009)
; 	  http://www.mathworks.com/matlabcentral/fileexchange/25449-image-edge-enhancing-coherence-filter-toolbox
; 	  2.  Nonlinear Diffusion Toolbox by Frederico D'Almeida (March 2004)
; 	  http://www.mathworks.com/matlabcentral/fileexchange/3710-nonlinear-diffusion-toolbox
; History: 
; 	    10 Feb 2011 - First version
;-
; Variable initialization
if n_params() lt 1 then $
message,'Please call the function imgdiff=nonlineardif (img,sigma, rho, m=m, dt=dt, ndt=ndt,mode=mode)'
if ~exist(sigma) then sigma=0.5 
if ~exist(rho) then rho=1.0
if ~exist(dt) then dt=0.01
if ~exist(ndt) then ndt=10
if ~exist(mode) then mode='eed'
if strcmp(mode,'eed') then rho=0. ; no smoothing for edge enhance filtering 
if ~exist(scheme) then scheme='standard'
if ~exist(u) then on_error,2
dif_time = 0
sz=size(u)
if sz[0] gt 2 then message,'Input image must be grayscale.'
y=double(u)

for i=1, ndt do begin
      grad=gsderiv(y, sigma, order=1)
      gradx=grad[*,*,0] & grady=grad[*,*,1] 
      SS=structuretensor(gradx,grady,rho)   ; Calculate structure tensor JJ[nx,ny,3]  
      eigen=eigenvector(SS[*,*,0],SS[*,*,1],SS[*,*,2]) ; Construct Eigenvalues and eigenvectors
;       mu1=reform(eigen.value[*,*,0]) & mu2= reform(eigen.value[*,*,1])
;       v1x=reform(eigen.vector[*,*,0]) & v1y=reform(eigen.vector[*,*,1])
;       v2x=reform(eigen.vector[*,*,2]) & v2y=reform(eigen.vector[*,*,3])
;'      help,eigen,/str
      DD=diffusiontensor(eigen,gradA=gradx^2+grady^2,mode=mode) 
      Dxx=DD[*,*,0] & Dxy=DD[*,*,1] & Dyy=DD[*,*,2]
;      help,DD
      case scheme of 
	      'standard':           y=standard(y,Dxx,Dxy,Dyy,dt) 
       'non_negative':           y=non_negative(y,Dxx,Dxy,Dyy,dt)
	         'implicit':           y=implicit(y,Dxx,Dxy,Dyy,dt)
	            'novel':           y=novel(y,Dxx,Dxy,Dyy,dt)
 'rotation_invariant':           y=rotation_invariant(y,Dxx,Dxy,Dyy,dt)
	              else:           message,'Unknown scheme'
      endcase
      dif_time = dif_time + dt    ; time step updating 
endfor
return,y
end